# This program goes through the arrays one by one. For each array, it first takes the corresponding
# Description file and loads it into a hash such that keys=accessions, values=descriptions. Then it
# takes the corresponding Blast results file and parses it using the Bioperl SearchIO module. 
# The details of the alignments like Query accession, Hit accession, definition lines,
# E value, etc are extracted and stored in a database if the Hit passes a given threshold E Value.
# This process is repeated for every array. In addition, the program creates a file 
# GeneSelectOutputInfo.txt giving a summary of the results.
#
#
# There must be two associated files *Description.txt and *Blast'Date'.txt for every array.
# The * is a placeholder for the Prefix, which corresponds to a given array. 
# The Description file has two tab delimited columns, the first column contains the
# Accession number and the second column contains the corresponding descrption provided
# by the array manufacturer.
# The Blast file is simply the Blast results file. The Blast results 
# of all the sequences represented on a given array are contained in a single file.
#
# The Prefix file lists the prefix for each array in a single column.
# 
#
# NOTE: This program was written on a WINDOWS 2000 machine.





#!/usr/bin/perl

use strict;
use warnings;

use DBI;			# This module is required to work with databases.
use DBD::ODBC;		# This module is the driver required for MS Access.
use Bio::SearchIO;	# Loads the SearchIO module of Bioperl required to parse Blast files.


my $DatabaseName = "GeneSelectDB";
# Holds the name of the Database or its alias.
# Make sure the Database is registered.

my $BlastSuffix = "Blast01112004\.txt";
# This is the suffix added to get the name of the Blast files given the prefix.
# I included the date the Blast was run.

my $DescriptionSuffix = "Description\.txt";
# This is the suffix added to get the name of the description files given the prefix.

my $EValueThreshold = "1e-015";
# This is the E Value threshold used to determine homology.
# The code forces Perl to use it as a float instead of as a string.


my $input;			# Used to store the PATH and name of the Prefix file, essentially a file 
					# containing a Prefix for every array. This permits batch
					# execution of all the arrays.
my $line;			# Used to read line by line from the Prefix file.
my $DataPath;		# Used to hold location of Blast and Description files.
my $OutputFile;		# Holds PATH and name of Output File, a file used to hold summary of results.
my $dbh;			# DB handler.
my $NumOfArray = 0;	# A counter to tell which array is currently being processed.


print "Please enter the name (including the PATH) of the file containg the prefixes:";
$input = <STDIN>;
chomp( $input );
$input =~ s/^\s+// if ( $input ); # get rid of leading spaces.
$input =~ s/\s+$// if ( $input ); # get rid of trailing spaces.


print "Please enter the location (PATH) of the Blast and the Desciption files:";
$DataPath = <STDIN>;
chomp( $DataPath );
$DataPath =~ s/^\s+// if ( $DataPath ); # get rid of leading spaces.
$DataPath =~ s/\s+$// if ( $DataPath ); # get rid of trailing spaces.


open( INPUT, "<$input" ) or die( "Couldn't open $input $! \n" );
# INPUT is the filehandle for the Prefix file.
# $input contains the name of the Prefix file.

$OutputFile = $DataPath . "GeneSelectOutputInfo\.txt";
open( OUTPUT, ">$OutputFile" ) or die( "Couldn't open $OutputFile $! \n" );
# OUTPUT is filehandle for the Output File.

# Print column names for the Output Summary file and to the console.
print( OUTPUT "NumOfArray\tPrefix\tNumOfGenes\tNumPassed\tNumUnique\tAvgLen\tAvgLenPass\n" );
print( "\nNumOfArray\tPrefix\tNumOfGenes\tNumPassed\tNumUnique\tAvgLen\tAvgLenPass\n" );


# Connect to the database.
$dbh = DBI->connect( "DBI:ODBC:$DatabaseName", "", "", { RaiseError => 1 } );
# Empty strings are username and password, none set on this machine.


# Start reading from the Prefix file to process the first array.
$line = <INPUT>;
chomp( $line );
$line =~ s/^\s+// if ( $line ); # get rid of leading spaces.
$line =~ s/\s+$// if ( $line ); # get rid of trailing spaces.


# This while loop goes through the Prefix file line by line and extracts data from 
# the Description file and Blast file for the array corresponding to the current prefix
# and enters it into the database.
while ( $line )
{
	
	my $BlastFile = $DataPath . $line . $BlastSuffix;
	my $DescriptionFile = $DataPath . $line . $DescriptionSuffix;
	my $object = new Bio::SearchIO( -fromat => 'blast', -file => $BlastFile );
	# The whole Blast results file is converted into an Object of type SearchIO class.

	my $result;			# Holds the Blast results for a given query.
	my @array;			# Holds results of the split function.
	my $DescLine;		# Used to read from the Description file.
	my %Description;	# Holds the Description file, keys=accessions, values=descriptions.
	my $NumOfGenes = 0;	# Counter to count the number of genes on each array.
	my $NumPassed = 0;	# Counter to count the number of genes that pass threshold E Value.
	my $NumUnique = 0;	# Counter to count the number of genes newly added to GenesNR.
	my $AvgLen = 0;		# Used to determine the average query sequence length.
	my $AvgLenPass = 0;	# Used to determine the avg query seq length for those that pass threshold.
	

	$NumOfArray++;		# Counts number of arrays.

	
	open( DESC, "<$DescriptionFile" ) or die( "Couldn't open $DescriptionFile $! \n" );
	# DESC is the filehandle for the Description file.

	$DescLine = <DESC>;
	chomp($DescLine) if ( $DescLine );
	$DescLine =~ s/^\s+// if ( $DescLine ); # get rid of leading spaces.
	$DescLine =~ s/\s+$// if ( $DescLine ); # get rid of trailing spaces.

	
	# This while loop goes through the Description file for the array corresponding to 
	# the current prefix and creates a hash such that keys=accessions, values=descriptions.
	while ( $DescLine )
	{
		
		@array = ();
		@array = split( /\t/, $DescLine );
		
		if (!($array[0]))
		{

			$array[0] = "No Value";
			# Somtimes an Accession may not exist, since some genes are listed as novel genes.
			# Setting it to some value ("No Value") will prevent bugs in the program.
			# This has no impact on the logic of the program.

		}
		if (!($array[1]))
		{
			
			$array[1] = "No Value";
			# Somtimes an Accession may not have any description associated with it.
			# Setting it to some value ("No Value") will prevent bugs in the program.
			# This has no impact on the logic of the program.

		}

		$array[0] = CleanAccession( $array[0] );
		$array[1] = CleanString( $array[1] );

		$Description{ $array[0] } = $array[1];

		$DescLine = <DESC>;
		chomp($DescLine) if ( $DescLine );
		$DescLine =~ s/^\s+// if ( $DescLine ); # get rid of leading spaces.
		$DescLine =~ s/\s+$// if ( $DescLine ); # get rid of trailing spaces.
		
	} # while ( $DescLine ) loop.

	close( DESC );

	$result = $object -> next_result;
	# gets the first result (Blast results of the first query) in the Blast file.

	# This while loop goes through the Blast file (query result by query result) for the array corresponding to 
	# the current prefix and parses out the requried information. Remember that The Blast results 
	# of all the sequences (queries) represented on the array are contained in this single file.
	while ( $result )
	{
		
		my $QueryAcc;		# Holds the Query Accession number.
		my $QueryDef;		# Holds the Query definition line.
		my $QueryDesc;		# Holds Query description from Description file.
		my $QueryLen;		# Holds Query sequence length.
		my $Hit;			# Holds the first hit of for a given query.
		my $HitAcc;			# Holds the Unigene SID of the Hit. 
		my $HitGBAcc;		# Holds the GenBank Accession number of the Hit.
		my $HitCID;			# Hols the Unigene Cluster ID of the Hit.
		my $HitDef;			# Holds the Definition line of the Hit.
		my $HitSig;			# Holds the E Value (significance) for the alignment.
		my $HSP;			# Holds the first HSP for the first Hit.
		my $HSPHitStrand;	# Holds the Strand of the Hit used for alignment.
		my $PlusMinus;		# Holds nature of alignment PlusPlus or PlusMinus.
		my $comment;		# Stores the source array information, the Prefix of the array.
				
		my $DBQuery;		# Used to query the database.
		my $sth;			# DB statement handler.
		my @row;			# Holds the results of a DB query.

		my @array;			# Temporary variable, holds results of split function.

		
		$NumOfGenes++;		# Counts number of genes on array.


		# I have to go a round about way to get the QueryAcc, because the SearchIO module
		# quite often gives the locus instead of the QueryAcc when using
		# $result -> query_accession.
		$QueryAcc = $result -> query_name;
		@array = split( /\|/, $QueryAcc );
		$QueryAcc = $array[3];
		$QueryAcc = CleanAccession( $QueryAcc );

		$QueryDef = $result -> query_description;
		$QueryDef = CleanString( $QueryDef );
		
		if ( !($QueryDesc = $Description{ $QueryAcc }) ) 
		{
			
			$QueryDesc = "NoValue";
			# Done to avoid runtime errors.

			
		}
		
		$QueryLen = $result -> query_length;

		$AvgLen = $AvgLen + $QueryLen;
		# Actullary keeps track of the running total length.
		# Is converted to an average at the end.

		$Hit = $result -> next_hit;
		# Gets the first hit for the current query.
		
		if ( $Hit )
		{	
			
			# Sometimes $Hit may not be defined. This way there will be no runtime errors.
			
			$HitSig = $Hit -> significance;
			
			$HitAcc = $Hit -> accession;
			# NOTE: Since the Blast was done against the Unigene database, the accession in $HitAcc 
			# is the Unigene sequence identifier, SID.
			$HitAcc = CleanAccession( $HitAcc );
			
			$HitDef = $Hit -> description;
			$HitDef = CleanString( $HitDef );
			
			$HitDef =~ m/\/gb=.*\/gi=/;
			# Will match the GenBank Accession Number of the Hit. The form of the Unigene Description 
			# line is: "AGENCOURT_11343981 NICHD_XGC_Tad2 Xenopus laevis cDNA clone IMAGE:6881639 5', 
			# mRNA sequence /clone=IMAGE:6881639 /clone_end=5' /gb=CB208784 /gi=29148784 /ug=Xl.24713
			# /len=926"
			
			$HitGBAcc = $&;
			# $& stores the string that was matched.
			$HitGBAcc =~ s/\/gb=//; # getting rid of the /gb= 
			$HitGBAcc =~ s/ \/gi=//; # getting rid of the /gi=
			$HitGBAcc = CleanAccession( $HitGBAcc );
			
			$HitDef =~ m/\/ug=.*\/len=/;
			# Will match the Unigene Cluster ID of the Hit.
			$HitCID = $&;
			# $& stores the string that was matched.
			$HitCID =~ s/\/ug=//; # getting rid of the /ug= 
			$HitCID =~ s/ \/len=//; # getting rid of the /len=
			$HitCID =~ s/^\s+// if ( $HitCID ); # get rid of leading spaces.
			$HitCID =~ s/\s+$// if ( $HitCID ); # get rid of trailing spaces.

			$HSP = $Hit -> next_hsp; # Gets the first HSP for the Hit.
			$HSPHitStrand = $HSP -> strand( 'hit' );
			if ( $HSPHitStrand eq "-1" )
			{
				$PlusMinus = "PlusMinus";
			}
			elsif ( $HSPHitStrand eq "1" )
			{
				$PlusMinus = "PlusPlus";
			}
			# Did the elsif instead of else eventhough the only values should be -1 and 1.
			# This way if something unexpected comes up $PlusMinus will be uninitialized
			# telling me that an error has occurred somewhere.
			
			$comment = $line;

		}
		else
		{
			
			# Since $Hit is not defined, $HitSig is set to an arbitrarily high value,
			# so that it will be ignored by the remainder of the program.
			$HitSig = "1000";

		}

		
		if ( $HitSig =~ m/^e/)
		{
			$HitSig = "9" . $HitSig;
		}
		# I am doing this because when the E Value is less than 1e-100, blastall.exe
		# does not give the base. I show an example of this below. So, I make it a 9, 
		# the highest integer it could be. This prevents bugs in the program and does not 
		# influence the logic of the program since our E Value threshold is far greater 
		# than 1e-100.
		#		
		# Query= gi|2219869|gb|AA490696.1|AA490696 aa45c03.r1 Soares_NhHMPu_S1
		# Homo sapiens cDNA clone IMAGE:823876 5' similar to gb:M60484_rna1
		# PROTEIN PHOSPHATASE PP2A-BETA, CATALYTIC SUBUNIT (HUMAN);, mRNA
		# sequence
		#		 (572 letters)
		#
		# Database: Xl.seq.uniq.Build46.txt 
		#		   22,135 sequences; 22,031,325 total letters
		#
		#
		#
		#																 Score    E
		# Sequences producing significant alignments:                      (bits) Value
		#
		# gnl|UG|Xl#S13590138 X.laevis mRNA for protein phosphatase 2A /cd...   450   e-126

		
		# Only Hits passing the threshold are entered in the database, so their significance is
		# compared to the threshold.
		if ( $HitSig < $EValueThreshold )
		{
			
			# The use of the '<' sign forces Perl to treat $HitSig and $EValueThreshold
			# as floating point numbers when evaluating this if statement.

			$NumPassed++;	# Counts number of genes that pass E Value threshold.

			$AvgLenPass = $AvgLenPass + $QueryLen;
			# Actullary keeps track of the running total length of queries that pass threshold.
			# Is converted to an average at the end.
			
			# quote() correctly quotes the strings in a way that is suitable 
			# for a given database engine, and thus prevents runtime errors.
			# If this is not done, pipe symbols, single quotes (Ex: 5'), etc. can
			# cause runtime errors.
			$QueryDef = $dbh->quote( $QueryDef );
			$QueryDesc = $dbh->quote( $QueryDesc );
			$HitDef = $dbh->quote( $HitDef );
						
			
			# prepare query.
			$DBQuery = "SELECT * FROM GenesNR WHERE Accession = '$HitGBAcc'";
			$sth = $dbh->prepare( $DBQuery ) 
					or die( "Cannot prepare $DBQuery: " . $dbh->errstr . "\n" );

			# execute query.
			$sth->execute
				or die( "Cannot execute query: " . $dbh->errstr . "\n" );

			@row = $sth->fetchrow_array; 
			# The the first row (if it exists) of the result is returned to an array.
			# Each field is one element of the array.

			if ($row[0])
			{
				
				# If $row[0] is defined, that means the gene is already in the GenesNR table (the 
				# nonredundant list of genes selected). So the entry will be added only to the Genes 
				# table that holds all entries, even duplicates and the comment column in the
				# GenesNR table will be updated to show that the gene was found on multiple arrays.
				
				$DBQuery = "INSERT INTO Genes ( SourceAccession, SourceDefinitionLine, SourceDescription, Accession, ClusterID, SID, DefinitionLine, Comment, EValue, PlusMinus )
							VALUES ( '$QueryAcc', $QueryDef, $QueryDesc, '$HitGBAcc', '$HitCID', '$HitAcc', $HitDef, '$comment', '$HitSig', '$PlusMinus' );";
				
				$dbh->do( $DBQuery );
				
				$comment = $row[7] . ", $comment"; # Updates the Comment column.

				$DBQuery = "UPDATE GenesNR SET Comment = '$comment' WHERE Accession = '$HitGBAcc'";
				$dbh->do( $DBQuery );

			}
			else
			{
				
				# If $row[0] is not defined, that means the gene is not in the database. 
				# So it will be added to both the GenesNR table (the nonredundant list of genes
				# selected) and the Genes table that holds all entries, even duplicates.
				
				$NumUnique++;	# Counts number of genes newly added to GenesNR from this array.
				
				$DBQuery = "INSERT INTO Genes ( SourceAccession, SourceDefinitionLine, SourceDescription, Accession, ClusterID, SID, DefinitionLine, Comment, EValue, PlusMinus )
							VALUES ( '$QueryAcc', $QueryDef, $QueryDesc, '$HitGBAcc', '$HitCID', '$HitAcc', $HitDef, '$comment', '$HitSig', '$PlusMinus' );";
				
				$dbh->do( $DBQuery );

				$DBQuery = "INSERT INTO GenesNR ( SourceAccession, SourceDefinitionLine, SourceDescription, Accession, ClusterID, SID, DefinitionLine, Comment, EValue, PlusMinus )
							VALUES ( '$QueryAcc', $QueryDef, $QueryDesc, '$HitGBAcc', '$HitCID', '$HitAcc', $HitDef, '$comment', '$HitSig', '$PlusMinus' );";
				
				$dbh->do( $DBQuery );
				

			} # if ($row[0]) else block.

		} # if ( $HitSig < $EValueThreshold ) block.

		$sth->finish() if ( $sth );
		
		$result = $object -> next_result;
		# Gets the Blast results of the next query.

	} # while ( $result ) loop.

	# The running total lengths are converted into averages.
	$AvgLen = $AvgLen/$NumOfGenes;
	$AvgLenPass = $AvgLenPass/$NumPassed;
	
	# Print summary to GeneSelectOutputInfo file.
	print( OUTPUT "$NumOfArray\t$line\t$NumOfGenes\t$NumPassed\t$NumUnique\t$AvgLen\t$AvgLenPass\n" );
	# Print to the cosole to keep user informed of progress.
	print( "$NumOfArray\t$line\t$NumOfGenes\t$NumPassed\t$NumUnique\t$AvgLen\t$AvgLenPass\n" );
	
	# Get the prefix of the next array.
	$line = <INPUT>;
	chomp( $line ) if ( $line );
	$line =~ s/^\s+// if ( $line ); # get rid of leading spaces.
	$line =~ s/\s+$// if ( $line ); # get rid of trailing spaces.

} # while ( $line ) loop.


$dbh->disconnect();
close( INPUT );
close( OUTPUT );


# This subroutine is used to get rid of the decimal portion of the accession number.
sub CleanAccession
{
	
	my $temp = $_[0];
	my @temp;
	
	@temp = split( /\./, $temp);
	$temp = $temp[0];
	$temp =~ s/^\s+// if ( $temp ); # get rid of leading spaces.
	$temp =~ s/\s+$// if ( $temp ); # get rid of trailing spaces.
	
	return $temp;

}


# This subrountine cleans up quotes, spaces, and semicolons in strings, 
# so they don't cause runtime errors when interacting with the database.
sub CleanString
{
	
	my $temp = $_[0];

	$temp =~ s/^\s+// if ( $temp ); # get rid of leading spaces.
	$temp =~ s/\s+$// if ( $temp ); # get rid of trailing spaces.
	$temp =~ s/\"//g;
	$temp =~ s/\;//g;
	
	return $temp;

}