# This program will download the Unigene records from NCBI's Unigene page for the (Xenopus) Accessions 
# in the GenesNR table based on the SID (i.e. the Unigene Sequence ID). Then it will parse the records  
# and store a summary of the records in the UnigeneRecordsSummary table. Afterwards, it will anlalyze  
# the summary and the results will be stored in the ResultsSummary table.
#
# Instead of completely relying one the Unigene data file, this script uses Unigene Records that are
# downloaded, because only they have the orientation information about the protein alignments that 
# are needed to determine the orientaion of the sequences. However, the Unigene data file is used to get
# the SEQTYPE of the sequence (3'EST, 5'EST, mRNA, HTC), since this information is not clearly marked
# in the downloaded Unigene records.
# The Unigene data file is Xx.data from ftp://ftp.ncbi.nih.gov/repository/UniGene/,
# where Xx is the species abbreviation (Xl for Xenopus laevis, Hs for Homo sapiens).
#
# NOTE: GeneSelect.pl must be run before using this program.
#
# The program starts by letting the user choose if he wants to download the Unigene Records or Parse and
# Analyze already downloaded records or all the above. The program is broken down into three subroutines
# DownloadRecords(), ParseRecords(), and AnlayzeRecords(). Their names indicate their functions.
#
# The DownloadRecords() subrountine is used to download the Unigene records. It does this by first getting 
# all the accessions found in the Accession column of the GenesNR table, and then constructs the required URL
# on the fly and downloads the Unigene record from the NCBI website using the LWP::UserAgent module.
# Using the downloaded HTML record and HTML::TokenParser module, a text file named 'accession'.txt
# is created for each accession number containing only the text portions of the HTML file. A sample file
# is pasted at the end of this script. The HTML version of the record ('accession'.html) is also saved for 
# future reference, since the online Unigene records are updated periodically. A log file called 
# DownloadErrorLog.txt is created and it lists the accessions for which Unigene records could not be 
# downloaded.
#
# The ParseRecords() subroutine is used to parse the downloaded Unigene records. It is used to parse
# text versions of the Unigene records. These files are of the form 'accession'.txt. A sample file is
# pasted at the end of the script. The program makes use the structure of the file to get the required
# information. It is the FEATURES information that is desired. Sequences can have three types of FEATURES
# CDS, prot_sim, and polyA_sig. The CDS FEATURE indicates whether the sequence has a designated CDS and if so,
# it also states which strand (+ or -) contians the CDS. The prot_sim FEATURE indicates any proteins sequences
# that align with the sequence in a translated blast search (BLASTX with -e 1e-6, see ProtEST website:
# http://www.ncbi.nlm.nih.gov/UniGene/protest.shtml). The BLASTX is done against 8 species, but we are 
# interested in only Human, Mouse, and Rat hits. The prot_sim FEATURE gives information about any protein
# hits and the strand (+ or -) used in the alignment. polyA_sig FEATURE indicates whether a polyA signal
# was detected and if so what strand (+ or -) it is on. It is the strand information we are primarily 
# inerested in, because it will be used by the AnalyzeRecords() subroutine to determine the orientation 
# of the sequences. The ParseRecords() subrountine starts by getting all the accession numbers from
# the GenesNR table and goes through the corresponding Unigene record text files one by one. The file
# is read line by line until the FEATURES section is found. Once the FEATURES section is reached, the
# program identifies which FEATURE it is dealing with and based on the structure of the FEATURE it
# extracts the necessary information. The information is stored in the UnigeneRecordsSummary table.
# In addition, the Unigene data file is used to get the SEQTYPE of the sequence (3'EST, 5'EST, mRNA,
# HTC), since this information is not clearly marked in the downloaded Unigene records.
# This process is repeated for all the accessions.
# It is important to note that some of the fields in the UnigneRecordsSummary table are not 
# applicable to every record.
# If an accession has no FEATURES associated with, it will not be found in the UnigeneRecordsSummary
# table.
#
# The AnalyzeRecords() subrountine is used to analyze the UnigeneRecordsSummary table. The subroutine
# gets all the accessions found in the Accession column of the GenesNR table and checks if it has any
# records associated with it in the UnigeneRecordsSummary table. If an accession does, the records are
# anlyzed to see if they are consistent with respect to orientation. For example, if an accession has
# 3 prot_sim records, the subrountine checks to see if all the prot_sim records relied on the same
# strand for the protein alignment. This anlaysis is done for every accession one by one.
#
#
# NOTE: This program was written on a WINDOWS 2000 machine.




#!/usr/bin/perl

use strict;
use warnings;

use DBI;				# This module is required to work with databases.
use DBD::ODBC;			# This module is the driver required for MS Access.
use LWP::UserAgent;		# This module is required to download files from the Internet.
use HTML::TokeParser;	# This module is used to parse HTML files.


my $DatabaseName = "GeneSelectDB";
# Holds the name of the Database or its alias.
# Make sure the Database is registered.

my $choice;			# Holds the choice entered by the user.
my $DataPath;		# Holds location of where the downloaded records should be stored.
my $DataFile;		# Holds the location of the Unigene Data file for a given Unigene Build. 
my $dbh;			# DB handler.
my $DBQuery;		# Used to query the database.
my $sth;			# DB statement handler.
my @row;			# Holds the results of a DB query.

print "Enter 1 to Download Unigene Records.\n";
print "Enter 2 to Parse and Analyze downloaded Unigene Records.\n";
print "Enter 3 to do all the above.\n";
print "Enter Choice: ";

# Read in the choice of user.
$choice = <STDIN>;
$choice =~ s/^\s+// if ( $choice ); # get rid of leading spaces.
$choice =~ s/\s+$// if ( $choice ); # get rid of trailing spaces.

# Connect to the database.
$dbh = DBI->connect( "DBI:ODBC:$DatabaseName", "", "", { RaiseError => 1 } );
# Empty strings are username and password, none set on this machine.

if ($choice == 1)
{

	print "Please enter the location (PATH) of where the downloaded records should be stored:";
	$DataPath = <STDIN>;
	chomp( $DataPath );
	$DataPath =~ s/^\s+// if ( $DataPath ); # get rid of leading spaces.
	$DataPath =~ s/\s+$// if ( $DataPath ); # get rid of trailing spaces.
	
	DownloadRecords();

}
elsif ($choice == 2)
{
	
	print "Please enter the location (PATH) of where the downloaded records are stored:";
	$DataPath = <STDIN>;
	chomp( $DataPath );
	$DataPath =~ s/^\s+// if ( $DataPath ); # get rid of leading spaces.
	$DataPath =~ s/\s+$// if ( $DataPath ); # get rid of trailing spaces.
	
	# The Unigene data file is needed to get the SEQTYPE of the sequence (3'EST, 5'EST, mRNA, HTC)
	# The SEQTYPE is not clearly marked in the downloaded Unigene Records.
	print "Please enter the name (including the PATH) of the Unigene Data file:";
	$DataFile = <STDIN>;
	chomp( $DataFile );
	$DataFile =~ s/^\s+// if ( $DataFile ); # get rid of leading spaces.
	$DataFile =~ s/\s+$// if ( $DataFile ); # get rid of trailing spaces.
	
	ParseRecords();
	AnalyzeRecords();

}
elsif ($choice == 3)
{
	
	print "Please enter the location (PATH) of where the downloaded records should be stored:";
	$DataPath = <STDIN>;
	chomp( $DataPath );
	$DataPath =~ s/^\s+// if ( $DataPath ); # get rid of leading spaces.
	$DataPath =~ s/\s+$// if ( $DataPath ); # get rid of trailing spaces.
	
	# The Unigene data file is needed to get the SEQTYPE of the sequence (3'EST, 5'EST, mRNA, HTC)
	# The SEQTYPE is not clearly marked in the downloaded Unigene Records.
	print "Please enter the name (including the PATH) of the Unigene Data file:";
	$DataFile = <STDIN>;
	chomp( $DataFile );
	$DataFile =~ s/^\s+// if ( $DataFile ); # get rid of leading spaces.
	$DataFile =~ s/\s+$// if ( $DataFile ); # get rid of trailing spaces.
	
	DownloadRecords();
	ParseRecords();
	AnalyzeRecords();

}
else
{

	print "\nYou did not enter a proper choice. Run the program again.\n";

}


$sth->finish() if ( $sth );
$dbh->disconnect();


sub DownloadRecords
{

	my $urlA;		# First part of URL from where the Unigene records will be downloaded.
	my $urlB;		# Second part of URL. Will hold the Unigene ID, the SID of the sequence (from GenesNR table).
	my $urlC = "&OPT=text";	# Last part of URL. Adds the text option to the URL, thus removing frames.
	my $ErrorLog = $DataPath . "DownloadErrorLog\.txt";
	# Holds the location of Error Log file.
	# The ErrorLog will list Accession for which records could not be downloaded.

	my $url;		# Holds the complete URL.
	
			
	open( LOG, ">$ErrorLog" ) or die( "Can not open $ErrorLog: $!" );;

	# prepare query.
	$DBQuery = "SELECT Accession, SID FROM GenesNR";
	$sth = $dbh->prepare( $DBQuery ) 
			or die( "Cannot prepare $DBQuery: " . $dbh->errstr . "\n" );

	# execute query.
	$sth->execute
		or die( "Cannot execute query: " . $dbh->errstr . "\n" );

	@row = $sth->fetchrow_array; 
	# The the first row (if it exists) of the result is returned to an array.
	# Each field is one element of array.

	
	# This while loop goes through all the accessions found in the GenesNR table one by one
	# and downloads the HTML Unigene Record for each accession based on the SID, the Unigene
	# Sequence identifier, and then parses the HTML file to get the text portions to create
	# the text file for each accession.
	while ( @row ) 
	{
		
		my $agent;		# Holds the LWP::UserAgent object.
		my $request;	# Holds the HTTP::Request object.
		my $response;	# Holds the Response to the HTTP request.
		my $document;	# Holds retrieved HTML page.

		my $page;		# Used by HTML::TokeParser.
		my $token;		# Used by HTML::TokeParser.
		my $type;		# Used by HTML::TokeParser.
		my $text;		# Used by HTML::TokeParser.
		
		my $file1;		# Holds the name of downloaded HTML file of Unigene Record.
		my $file2;		# Holds the name of the created text file of Unigene Record.
		my $SID;		# Holds the Unigene SID of the sequence.
		my $species;	# Holds the spcecies abbreviation.
		

		$SID = $row[1];
		# Get rid of the prefix of the SID.
		$SID =~ m/\#S/;
		$SID = $';
		# $' stores the string following what was matched.

		$species = $`;
		# $` stores the string preceding what was matched.
		# In this instance it is the species abbreviation.

		$urlB = $SID;

		$urlA = "http://www.ncbi.nlm.nih.gov/UniGene/seq.cgi?ORG=" . $species . "&SID=";
		
		# Create the working URL by combining all the parts.
		$url = $urlA . $urlB . $urlC;

				
		$agent = new LWP::UserAgent();
		$request = new HTTP::Request( 'GET' => $url );
		$response = $agent -> request( $request );

		# The HTML file is stored so that it can be used for future reference.
		# The text file, creted by parsing out all the  text portions of the
		# HTML file will be used later to parse out specifc information.
		$file1 = $DataPath . "HTML\\" . $row[0] . "\.html";
		$file2 = $DataPath . "TXT\\" . $row[0] . "\.txt";
		
		open( OUTPUT1, ">$file1" ) or die( "Can not open $file1: $!" );
		open( OUTPUT2, ">$file2" ) or die( "Can not open $file2: $!" );

		if ( $response -> is_success )
		{
			$document = $response -> content();
			print( OUTPUT1 "$document" );
			
			$page = new HTML::TokeParser( \$document );

			# This loop prints only the text portions of the HTML file.
			while ( $token = $page -> get_token() )
			{
				$type = shift( @{ $token } );
				$text = shift( @{ $token } );

				if ( $type eq "T" )
				{
					print( OUTPUT2 "$text" );
				}

			}

			print "Downloaded $row[0]\n";
			
		}
		else
		{
			print( LOG "Error in $row[0]\t$row[1]: " . $response -> status_line() . "\n" );
		}

		close( OUTPUT1 );
		close( OUTPUT2 );

		@row = $sth->fetchrow_array; 
		# the the next row is returned to an array.

			
	} # while ( @row ) loop.

	close( LOG );

	print "\nFinished DownloadRecords()\n\n";

} # sub DownloadRecords block.


sub ParseRecords
{

	my %SeqType;			# Hash to hold the SeqTypes (3'EST, 5'EST, mRNA, HTC).
	my $key;				# Used to hold the key of the hash entry.
	my $value;				# Used to hold the value of the hash entry.
	my $data;				# Used to read the DataFile line by line.
	my $temp;				# A temporary variable.
	
		
	# First the Unigene data file is opened.
	open( DATA, "<$DataFile" ) or die( "Cannnot open $DataFile: $!" );

	$data =  <DATA>;

	# This while loop will go through the Unigene datafile line by line and for every accession number
	# it will extract the sequence type (SeqType) and construct a hash such that the key=accession
	# and value=SeqType. This hash will be used later in the subrountine. In the Unigene datafile, 
	# any line giving sequence information starts with the word SEQUENCE. The program takes adavantage
	# of this file structure. Also, the SEQTYPE is usually the last part of the line and is clearly 
	# idendified as SEQTYPE=. When the SEQTYPE=EST, we need to know if it is a 3' EST or 5' EST. 
	# This information is given by the END= keyword. Bascially, the Unigene data file has a 
	# well-defined structre that the program uses to get the required information.
	while ( $data )
	{
		
		if ( $data =~ m/^SEQUENCE/ )
		{
			
			$data =~ m/ACC=.*\; NID=/; # mathes the accession number.
			$key = $&;
			# $& stores the string that was matched.
			$key =~ s/ACC=//; # getting rid of the 'ACC='
			$key =~ s/\; NID=//; # getting rid of the '; NID='
			$key = CleanAccession( $key );

			$data =~ m/SEQTYPE=.*\n/; # matches the SeqType.
			$value = $&;
			# $& stores the string that was matched.
			$value =~ s/SEQTYPE=//; # getting rid of the 'SEQTYPE='
			chomp( $value );

			if ( $value =~m/TRACE/ )
			{
				# This means that the SEQTYPE is EST and there is additional
				# information. All that is needed is that SEQTYPE=EST,
				# so $value will be set to EST.

				$value = "EST";
			}

			# If value=EST, we need get further information
			# about which END the EST is from 3' or 5'.
			# This information is given by the END keyword.
			# Some ESTs may not have END information.
			if ( $value eq "EST" and $data =~ m/END=/ )
			{
				
				$data =~ m/END=.*\; LID=/;
				$temp = $&;
				# $& stores the string that was matched.
				$temp =~ s/END=//; # getting rid of the 'END='
				$temp =~ s/\; LID=//; # getting rid of the '; LID='
				$value = $temp . $value;

			}

			# Enter the accession and SeqType into the hash.
			$SeqType{ $key } = $value;

		}

		$data =  <DATA>;

	} # while( $data ) loop.

	close( DATA );


	# Get all the accessions in GenesNR table.
	# prepare query.
	$DBQuery = "SELECT Accession FROM GenesNR";
	$sth = $dbh->prepare( $DBQuery ) 
			or die( "Cannot prepare $DBQuery: " . $dbh->errstr . "\n" );

	# execute query.
	$sth->execute
		or die( "Cannot execute query: " . $dbh->errstr . "\n" );

	@row = $sth->fetchrow_array; 
	# The the first row (if it exists) of the result is returned to an array.
	# Each field is one element of the array.

	
	# This while loop goes through all the accession and the associated Unigene record text files
	# one by one and extracts all the requried information and enters it into the 
	# UnigeneRecordsSummary table.
	while ( @row ) 
	{
		
		my $file = $DataPath . "TXT\\" . $row[0] . "\.txt";
		# Holds loaction of Unigene record text file for the current accession.

		my $Accession;				# Holds the accession.
		my $ProtAcc = "";			# Holds the protein accession.
		my $ProtDesc = "";			# Holds protein description.
		my $organism = "";			# Holds name of organism.
		my $EValue;					# Holds the EValue of alignment.
		my $SeqType = "";			# Holds the sequence type, i.e. mRNA, 3'EST, 5'EST.
		my $ProtAlign = "";			# Holds strand used for alignment.
		my $CDS = "";				# Holds which strand CDS is on.
		my $PolyASignal = "";		# Holds which strand PolyA signal is on.
		my $line;					# Used to read from $file line by line.
		my @array;					# A temporary variable.
		my @temp;					# A temporary variable.

		
		$Accession = $row[0]; # Set the accession number.
		$SeqType = $SeqType{ $Accession }; # Gets the SeqType from the hash.
		$SeqType = $dbh -> quote( $SeqType ); # Quotes the SeqType properly for database entry.
		
		# Open Unigene record text file for the currnet accession.
		open( FILE, "<$file" ) or die( "Cannnot open $file: $!" );

		$line = <FILE>;
		
		# This while loop gets to the FEATURES section of the Unigene record.
		while ( $line and $line ne "FEATURES\n" )
		{
			$line = <FILE>;
		}

		$line = <FILE>;
		
		# This while loop identifies which FEATURE we are dealing with(CDS, prot_sim, polyA_sig)
		# gets the required information by making use of the structure of that FEATURE. The
		# extracted information is entered into the UnigeneRecordsSummary table.
		# It is important to note that some of the fields in the UnigneRecordsSummary table are not 
		# applicable to every record.
		while ( $line )
		{

			# This if block identifies the prot_sim FEATURE and uses its
			# structure to get the required information.
			if ( $line eq "prot_sim\n" )
			{
	
				@array = ();
				$array[0] = <FILE>;
				$array[1] = <FILE>;
				$array[2] = <FILE>;
				$array[3] = <FILE>;

				# We will consider only Human, Mouse, or Rat alignments.
				if ( $array[1] =~ m/^(Homo|Mus|Rattus)/ )
				{
					
					# This if block gets the strand used for the alignment.
					if ( substr( $array[0], 1, 1) eq "+" )
					{
						$ProtAlign = "Plus";
					}
					elsif ( substr( $array[0], 1, 1) eq "-" )
					{
						$ProtAlign = "Minus";
					}

					chomp( $array[1] );
					@temp = split( /\t/, $array[1] );

					$organism = $temp[0];
					
					$ProtDesc = CleanString( $temp[1] );
					# quote() correctly quotes the strings in a way that is suitable 
					# for a given database engine, and thus prevents runtime errors.
					# If this is not done, pipe symbols, single quotes (Ex: 5'), etc. can
					# cause runtime errors.
					$ProtDesc = $dbh -> quote( $ProtDesc );

					# Gets the EValue for the alignment.
					# NOTE: The Unigene records incorreclty refer to it as a
					# p value, though it is an E Value. Personal communication 
					# with the NCBI staff has confirmed this.
					$array[3] =~ m/p=.*\)/i;
					$EValue = $&;
					# $& stores the string that was matched.
					$EValue =~ s/p=//; # getting rid of the 'p=' in the beginning. 
					$EValue =~ s/\)//; # getting rid of the ')' at the end.
					if ( $EValue =~ m/^e/)
					{
						$EValue = "9" . $EValue;
					}
					# I am doing this because sometimes the E Value does not have a base.
					# So, I make it a 9, the highest integer it could be. This prevents bugs 
					# in the program and does not influence the logic of the program. 

					# Get the protein accession for sequence used in alignment.
					$array[3] =~ m/\:.*Also/;
					$ProtAcc = $&;
					# $& stores the string that was matched.
					$ProtAcc =~ s/\://; # getting rid of the ':' in the beginning. 
					$ProtAcc =~ s/Also//; # getting rid of the 'Also' at the end.
					$ProtAcc = CleanAccession( $ProtAcc );

					# Prepare Query to enter data into table. Make all the fields not applicable to
					# a prot_sim record empty strings.
					$DBQuery = "INSERT INTO UnigeneRecordsSummary ( Accession, ProteinAccession, ProteinDescription, Organism, EValue, SeqType, ProteinAlignment, CDS, PolyAsignal )
							VALUES ( '$Accession', '$ProtAcc', $ProtDesc, '$organism', '$EValue', $SeqType, '$ProtAlign', '', '' );";
				
					$dbh->do( $DBQuery );
					
				} # if ( $array[1] =~ m/^(Homo|Mus|Rattus)/ ) block.

			} # if ( $line eq "prot_sim\n" ) block.
			# The elsif block below identifies the CDS FEATURE and uses its
			# structure to get the required information.
			elsif ( $line eq "CDS\n" )
			{

				@array = ();
				$array[0] = <FILE>;
				$array[1] = <FILE>;

				# This if block determines which strand the CDS is on.
				if ( substr( $array[0], 1, 1) eq "+" )
				{
					$CDS = "Plus";
				}
				elsif ( substr( $array[0], 1, 1) eq "-" )
				{
					$CDS = "Minus";
				}

				# Get the description about the prodcut of the CDS.
				chomp( $array[1] );
				$ProtDesc = $array[1];
				$ProtDesc = CleanString( $array[1] );
				# quote() correctly quotes the strings in a way that is suitable 
				# for a given database engine, and thus prevents runtime errors.
				# If this is not done, pipe symbols, single quotes (Ex: 5'), etc. can
				# cause runtime errors.
				$ProtDesc = $dbh->quote( $ProtDesc );

				# Prepare Query to enter data into table. Make all the fields not applicable to
				# a CDS record empty strings.
				$DBQuery = "INSERT INTO UnigeneRecordsSummary ( Accession, ProteinDescription, SeqType, CDS, ProteinAccession, Organism, ProteinAlignment, PolyASignal )
							VALUES ( '$Accession', $ProtDesc, $SeqType, '$CDS', '', '', '', '' );";
				
				$dbh->do( $DBQuery );
				
			} # elsif ( $line eq "CDS\n" ) block.
			# The elsif block below identifies the oplyA_sig FEATURE and uses its
			# structure to get the required information.
			elsif ( $line eq "polyA_sig\n" )
			{
				
				@array = ();
				$array[0] = <FILE>;

				# This if block determines which strand the PolyA signal is on.
				if ( substr( $array[0], 1, 1) eq "+" )
				{
					$PolyASignal = "Plus";
				}
				elsif ( substr( $array[0], 1, 1) eq "-" )
				{
					$PolyASignal = "Minus";
				}

				# Prepare Query to enter data into table. Make all the fields not applicable to
				# a polyA_sig record empty strings.
				$DBQuery = "INSERT INTO UnigeneRecordsSummary ( Accession, SeqType, PolyASignal, ProteinAccession, ProteinDescription, Organism, ProteinAlignment, CDS )
							VALUES ( '$Accession', $SeqType, '$PolyASignal', '', '', '', '', '' );";
				
				$dbh->do( $DBQuery );

			} # elsif ( $line eq "polyA_sig\n" ) block.

			$line = <FILE>;

		} # while ( $line ) loop.
		
		# close the Unigene record for the current accession.
		close( FILE );
		
		# Get the next accession.
		@row = $sth->fetchrow_array; 
		# the the next row is returned to an array.

		print "Parsed $row[0]\n";
		
	} # while ( @row ) loop.

	print "\nFinished ParseRecords()\n\n";

} # sub ParseRecords block.


sub AnalyzeRecords
{
	
	# The first part of AnalyzeRecords is the same as ParseRecords, because we need to make
	# %SeqType. The %SeqType hash will be used to determine the SeqType for the accessions
	# not found in the UnigeneResultsSummary table. Remember, if an accession has no FEATURES 
	# associated with it, then it will not be found in the UnigeneRecordsSummary table.
	# This code cannot be readily put in a subroutine, because a subroutine cannot return a hash.
	
	my %SeqType;			# Hash to hold the SeqTypes (3'EST, 5'EST, mRNA, HTC).
	my $key;				# Used to hold the key of the hash entry.
	my $value;				# Used to hold the value of the hash entry.
	my $data;				# Used to read the DataFile line by line.
	my $temp;				# A temporary variable.

	my $QUERY;				# Used to query the database.
	my $STH;				# A statement handler.
	
		
	# First the Unigene data file is opened.
	open( DATA, "<$DataFile" ) or die( "Cannnot open $DataFile: $!" );

	$data =  <DATA>;

	# This while loop will go through the Unigene datafile line by line and for every accession number
	# it will extract the sequence type (SeqType) and construct a hash such that the key=accession
	# and value=SeqType. This hash will be used later in the subrountine. In the Unigene datafile, 
	# any line giving sequence information starts with the word SEQUENCE. The program takes adavantage
	# of this file structure. Also, the SEQTYPE is usually the last part of the line and is clearly 
	# idendified as SEQTYPE=. When the SEQTYPE=EST, we need to know if it is a 3' EST or 5' EST. 
	# This information is given by the END= keyword. Bascially, the Unigene datafile has a 
	# well-defined structre that the program uses to get the required information.
	while ( $data )
	{
		
		if ( $data =~ m/^SEQUENCE/ )
		{
			
			# Get the accession.
			$data =~ m/ACC=.*\; NID=/; # mathes the accession number.
			$key = $&;
			# $& stores the string that was matched.
			$key =~ s/ACC=//; # getting rid of the 'ACC='
			$key =~ s/\; NID=//; # getting rid of the '; NID='
			$key = CleanAccession( $key );

			# Get the SEQTYPE.
			$data =~ m/SEQTYPE=.*\n/; # matches the SeqType.
			$value = $&;
			# $& stores the string that was matched.
			$value =~ s/SEQTYPE=//; # getting rid of the 'SEQTYPE='
			chomp( $value );

			if ( $value =~m/TRACE/ )
			{
				# This means that the SEQTYPE is EST and there is additional
				# information. All that is needed is that SEQTYPE=EST,
				# so $value will be set to EST.

				$value = "EST";
			}

			# If value=EST, we need get further information
			# about which END the EST is from 3' or 5'.
			# This information is given by the END keyword.
			# Some ESTs may not have END information.
			if ( $value eq "EST" and $data =~ m/END=/ )
			{
				
				$data =~ m/END=.*\; LID=/;
				$temp = $&;
				# $& stores the string that was matched.
				$temp =~ s/END=//; # getting rid of the 'END='
				$temp =~ s/\; LID=//; # getting rid of the '; LID='
				$value = $temp . $value;

			}

			# Enter the accession and SeqType into the hash.
			$SeqType{ $key } = $value;

		} # if ( $data =~ m/^SEQUENCE/ ) block.

		$data =  <DATA>;

	} # while( $data ) loop.

	close( DATA );
	
	
	# Get all the accessions in GenesNR table.
	# prepare query.
	$QUERY = "SELECT Accession, PlusMinus FROM GenesNR ORDER BY Accession ASC";
	$STH = $dbh->prepare( $QUERY ) 
			or die( "Cannot prepare $QUERY: " . $dbh->errstr . "\n" );

	# execute query.
	$STH->execute
		or die( "Cannot execute query: " . $dbh->errstr . "\n" );

	@row = $STH->fetchrow_array; 
	# The the first row (if it exists) of the result is returned to an array.
	# Each field is one element of the array.

	while ( @row )
	{
		
		my $Accession;			# Holds the accession.
		my $result;				# Holds the result of a query to the UnigeneRecordsSummary table.
		my @record;				# Holds a record, each element corresponds to a field.
		my $SeqType = "";		# Holds the sequence type, i.e. mRNA, 3'EST, 5'EST.
		my $PolyASignal = "";	# Holds which strand PolyA signal is on.
		my $Features = "";		# Indicates whether the accession has any FEATURES.
		my $Consistency = "";	# Indicates whether the records of a given accession are consistent with respect to orientaiton,
		my $Orientation = "";	# Indicates the orientation, if all the records of a given accession agree.
		my $PlusMinus = "";		# It is the PlusMinus value from the GenesNR table.
		my $NumRecords;			# The number of records a given accession has in the UnigeneRecordsSummary table.
		my $PolyASig = "";		# Yes if polyA_sig is present, otherwise No.
		my $temp;				# A temporary variable.
		
	
		$Accession = $row[0];
		$PlusMinus = $row[1];

		# We must check if this accession is present in the UnigeneRecordsSummary table.
		# Remember, if an accession has no FEATURES associated with it, it will not be found 
		# in the UnigeneRecordsSummary table.
		# prepare query.
		$DBQuery = "SELECT * FROM UnigeneRecordsSummary WHERE Accession = '$Accession' ORDER BY PolyASignal DESC";
		$sth = $dbh->prepare( $DBQuery ) 
				or die( "Cannot prepare $DBQuery: " . $dbh->errstr . "\n" );
		# Because the Query orders the results by PolyASignal in a Descending order, if that accession has a
		# polyA_sig record, it will be the first record, since the PolyASignal field will be an empty string
		# for non polyA_sig records.

		# execute query.
		$sth->execute
			or die( "Cannot execute query: " . $dbh->errstr . "\n" );

		$result = $sth->fetchall_arrayref; 
		# All the results of the Query are fetched at once.
		# This returns a reference to an array containing references to each row
		# of the result.


		if ( $$result[0] )
		{
				# If the result is defined, then we check if only one or multiple records
				# are returned.
				# If thre is only one record for the accession in UnigeneRecordsSummary,
				# we need to check if it is a polyA_sig record. 
				if ( scalar(@$result) == 1 )
				{
					$temp = $$result[0];
					@record = @$temp;
					
					# $record[8] is the PolyASignal field.
					# if  $record[8] is an empty string, that means the record is not
					# a polyA_sig record. if it is not an empty string, then it is a
					# polyA_sig record.
					if ( $record[8] ne "" )
					{
						$Features = "OnlyPolyASignal";
						$NumRecords = 1;
						$SeqType = $record[5];
						$SeqType = $dbh -> quote( $SeqType ); # Quotes the SeqType properly for database entry.
						$Orientation = $record[8];
						$PolyASig = "Yes";

						# prepare query to enter data into table.
						$DBQuery = "INSERT INTO ResultsSummary ( Accession, Features, NumberOfRecords, Orientation, SeqType, PlusMinus, PolyASig, Consistency )
									VALUES ( '$Accession', '$Features', '$NumRecords', '$Orientation', $SeqType, '$PlusMinus', '$PolyASig', '' );";
						
						$dbh->do( $DBQuery );

					}
					else
					{
						$Features = "Present";
						$NumRecords = 1;
						$SeqType = $record[5];
						$SeqType = $dbh -> quote( $SeqType ); # Quotes the SeqType properly for database entry.
						$PolyASig = "No";
						$Consistency = "Consistent";
						
						# Since it is not a polyA_sig record, it has to be either a prot_sim record
						# or a CDS record. If it is a prot_sim record the orientaion is given by $record[6].
						# If it is a CDS record the orientaion is given by $record[7].
						# This if block gets the orientation from the appropriate place.
						if ( $record[6] ne "" )
						{
							$Orientation = $record[6];
						}
						else
						{
							$Orientation = $record[7];
						}

						# prepare query to enter data into table.
						$DBQuery = "INSERT INTO ResultsSummary ( Accession, Features, NumberOfRecords, Consistency, Orientation, SeqType, PlusMinus, PolyASig )
									VALUES ( '$Accession', '$Features', '$NumRecords', '$Consistency', '$Orientation', $SeqType, '$PlusMinus', '$PolyASig' );";
						
						$dbh->do( $DBQuery );

					}

				} # if ( scalar(@$result) == 1 ) block.
				else
				{
					my $i;					# Temporary variable.
					my $m;					# Holds the index of the first non polyA_sig record.
					my $n;					# Gives the final index of @$result.
					my $CurrentOrientation;	# Holds orientaion of current record.

					# Get the first record. It will be used to get the SEQTYPE.
					$temp = $$result[0];
					@record = @$temp;
					
					$Features = "Present";
					$NumRecords = scalar( @$result );
					$SeqType = $record[5];
					$SeqType = $dbh -> quote( $SeqType ); # Quotes the SeqType properly for database entry.
					
					
					# $record[8] is the PolyASignal field.
					# if  $record[8] is an empty string, that means the record is not
					# a polyA_sig record. if it is not an empty string, then it is a
					# polyA_sig record.
					if ( $record[8] ne "" )
					{
						$PolyASig ="Yes";
						$PolyASignal = $record[8];
						# $PloyASignal holds which strand the PolyA signal is on.
					}
					else
					{
						$PolyASig ="No";
					}

					# $m holds the index of the first non polyA_sig record.
					if ( $PolyASig eq "Yes" )
					{
						$m = 1;
					}
					else
					{
						$m = 0;
					}
					
					$n = scalar( @$result ) - 1;

					$temp = $$result[$m];
					@record = @$temp;
					
					# Since it is not a polyA_sig record, it has to be either a prot_sim record
					# or a CDS record. If it is a prot_sim record the orientaion is given by $record[6].
					# If it is a CDS record the orientaion is given by $record[7].
					# This if block gets the orientation from the appropriate place.
					if ( $record[6] ne "" )
					{
						$Orientation = $record[6];
					}
					else
					{
						$Orientation = $record[7];
					}

					# This foreach loop will determine whether all the non polyA_sig records are
					# consistent with one another with respect to orientaiton.
					# The loop gets the orientaion of the current record and compares it with the
					# orientation of the first non polyA_sig record. Any mismatch indicates an
					# inconsistency. If the orientation of all the (non polyA_sig) records is equal to
					# the orientation of the first non polyA_sig record, then all the (non ployA_sig)
					# records are consistent with respect to orientation. If they are all consistent,
					# their orientation is compared to the orientation of the PolyA signal. The polyA
					# signal is treated separately since it is only a few bases in length.
					foreach $i ($m..$n)
					{
						$temp = $$result[$i];
						@record = @$temp;
						
						# The if block below gets the orientation of the current record.
						# Since it is not a polyA_sig record, it has to be either a prot_sim record
						# or a CDS record. If it is a prot_sim record the orientaion is given by $record[6].
						# If it is a CDS record the orientaion is given by $record[7].
						# This if block gets the orientation from the appropriate place.
						if ( $record[6] ne "" )
						{
							$CurrentOrientation = $record[6];
						}
						else
						{
							$CurrentOrientation = $record[7];
						}

						# Compare the orientation of current record with that of the first
						# non polyA_sig record. If even one record is not consistent, we 
						# exit the foreach loop.
						if ( $CurrentOrientation eq $Orientation )
						{
							$Consistency = "Consistent";
						}
						else
						{
							$Consistency = "NotConsistent";
							last;
						}
					}

					# if all the non polyA_sig records are consistent and a polyA_sig record
					# exists for the current accession, the orientation of the polyA signal
					# is considered.
					if ( $Consistency eq "Consistent" )
					{
						if ( ($PolyASig eq "Yes") and ($Orientation eq $PolyASignal) )
						{
							$Consistency = "ConsistentWithPolyA";
						}
					}
					else
					{
						$Orientation = "";
					}

					# prepare query to enter data into table.
					$DBQuery = "INSERT INTO ResultsSummary ( Accession, Features, NumberOfRecords, Consistency, Orientation, SeqType, PlusMinus, PolyASig )
								VALUES ( '$Accession', '$Features', '$NumRecords', '$Consistency', '$Orientation', $SeqType, '$PlusMinus', '$PolyASig' );";
						
					$dbh->do( $DBQuery );

				} # else part of if ( scalar(@$result) == 1 ) block.

		} # if ( $$result[0] ) block.
		else
		{
			$Features = "Absent";
			$NumRecords = 0;
			$SeqType = $SeqType{ $Accession }; # Gets the SeqType from the hash.
			$SeqType = $dbh -> quote( $SeqType ); # Quotes the SeqType properly for database entry.

			# prepare query to enter data into table.
			$DBQuery = "INSERT INTO ResultsSummary ( Accession, Features, NumberOfRecords, SeqType, PlusMinus, Consistency, Orientation, PolyASig )
						VALUES ( '$Accession', '$Features', '$NumRecords', $SeqType, '$PlusMinus', '', '', '' );";
			
			$dbh->do( $DBQuery );

		} # else part of if ( $$result[0] ) block.
		

		@row = $STH->fetchrow_array; 
		# Gets the next row if it exists.
		# Each field is one element of the array.

		print "Analyzed $row[0]\n";
	
	} # while ( @row ) loop.

	print "\nFinished AnalyzeRecords()\n\n";

} # sub AnalyzeRecords block.










# This subroutine is used to get rid of the decimal portion of the accession number.
sub CleanAccession
{
	my $temp = $_[0];
	my @temp;
	
	@temp = split( /\./, $temp);
	$temp = $temp[0];
	$temp =~ s/^\s+// if ( $temp ); # get rid of leading spaces.
	$temp =~ s/\s+$// if ( $temp ); # get rid of trailing spaces.
	
	return $temp;
}


# This subrountine cleans up quotes, spaces, and semicolons in strings, 
# so they don't cause bugs when interacting with the Database.
sub CleanString
{
	my $temp = $_[0];

	$temp =~ s/^\s+// if ( $temp ); # get rid of leading spaces.
	$temp =~ s/\s+$// if ( $temp ); # get rid of trailing spaces.
	$temp =~ s/\"//g; # get rid of double quotes.
	$temp =~ s/\;//g; # get rid of semicolon.
	
	return $temp;
}




#Sample Unigene Record (only the text portions of http://www.ncbi.nlm.nih.gov/UniGene/seq.cgi?ORG=Xl&SID=13807647&OPT=text)
#
# NCBI UniGene  
#
# Xl#S13807647 
# 3955 bases      
# Xenopus laevis     
# UniGene Xl.23131
# 
# Xenopus laevis mRNA for receptor tyrosine kinase Xror2, complete cds
# 
# 
# SEQUENCE INFORMATION
# 
#  
#   
#   GenBank entry:
# AB087137.1
#  
# 
# 
#  
#   
#   Product:
# PID g22831326
#  - receptor tyrosine kinase Xror2
#  
# 
# 
#  
#   
#   Sequence length:
# 3955 bases
#  
# 
# 
#  
#   
#   WARNING:
# Contains repetitive sequences
#  
# 
#  
# 
# FEATURES
# 
# 
# 
# CDS
# (+)145 ... 2937
# Product: receptor tyrosine kinase Xror2See protein sequence PID:g22831326
# 
# 
# prot_sim
# (+)205 ... 2931
# Homo sapiens	ROR2_HUMAN Tyrosine-protein kinase transmembrane receptor ROR2 precursor (Neurotrophic tyrosine kinase, receptor-related 2)
#  
# (80% identity in aligned region, p=0.0)See protein sequence sp:Q01974Also See: ProtEST
# 
# 
# prot_sim
# (+)181 ... 2931
# Mus musculus	ROR2_MOUSE Tyrosine-protein kinase transmembrane receptor ROR2 precursor (Neurotrophic tyrosine kinase, receptor-related 2) (mROR2)
#  
# (78% identity in aligned region, p=0.0)See protein sequence sp:Q9Z138Also See: ProtEST
# 
# 
# prot_sim
# (+)1546 ... 2376
# Rattus norvegicus	 muscle-specific tyrosine kinase receptor MuSK [Rattus norvegicus]
#  
# (41% identity in aligned region, p=6.3e-65)See protein sequence ref:NP_112323.1Also See: ProtEST
# 
# 
# prot_sim
# (+)643 ... 2415
# Caenorhabditis elegans	T18840 hypothetical protein C01G6.8 - Caenorhabditis elegans
#  
# (31% identity in aligned region, p=2.2e-94)See protein sequence pir:T18840Also See: ProtEST
# 
# 
# prot_sim
# (+)877 ... 2382
# Drosophila melanogaster	 Ror-P1 [Drosophila melanogaster]
#  
# (40% identity in aligned region, p=2e-108)See protein sequence ref:NP_476962.1Also See: ProtEST
# 
# 
# polyA_sig
# (+)3905 ... 3910
# 
# 
# 
# 
# return to NCBI mode
# 
# 